(* ########################################################################## *)
(* ############################### bsmlutils  ############################### *)
(* ########################################################################## *)

(** Some little expressions which are used in BSML programs *)

Require PolyList.
Require ZArith.
Require Omega.
Require Zwf.
Require Bsmllemma.

(** ** From "bsmlutils.ml" of the BSMLlib library *)

(* ************************************************************************** *)
(* ************************** like Bsmllib ********************************** *)
(* ************************************************************************** *)


(** Construct a list singleton *)
(* ****************************** mklist ************************************ *)
Definition mklist := [T:Set][x:T] (cons x (nil T)).


(** The identity *)
(* ****************************** Identity ********************************** *)
Definition id := [T:Set][x:T] x.

(** The composition *)
(* ****************************** Composition ******************************* *)
Definition compose := [T1,T2,T3:Set][f:T2->T3][g:T1->T2][x:T1](f (g x)).


(** Decidability of the "Option" *)
Definition option_dec:(T:Set) (arg:(option T)) {x:T | arg=(Some T x)}+{arg=(None T)}.
Intros.
Case arg;Intros;Auto.
Left.
Exists t;Trivial.
Defined.


(** Test if an "'a option" is empty or not *)
Definition is_Some_def [T:Set][x:(option T)] : Prop := 
   Cases x of (Some _) => True | None => False end.

Definition is_Some : (T:Set)(arg:(option T)){is_Some_def T arg}+{~(is_Some_def T arg)}.
Intros. 
Case arg;Intros;Simpl.
Left;Trivial.
Right;Auto.
Defined.


(* ************************************************************************** *)
(* ****************************** Z List ************************************ *)
(* ************************************************************************** *)

(** ** Tools on lists 
 functions and inductives to manipulates polymorphic list with "Z" (ZArith) instead of "N" (Arith) *)



(* ************************************************************************** *)
(* ****************************** Zlenght *********************************** *)
(* ************************************************************************** *)

(** Length of a list *)
Fixpoint Zlength [T:Set] [l:(list T)] : Z := Cases l of 
                                     nil => `0` 
		            | (cons _ m) => `(Zlength T m)+1` end.
			    

Lemma Zlength_app : (T:Set) (l1,l2:(list T)) ( (Zlength T (app l1 l2))=`(Zlength T l1)+(Zlength T l2)`).
Intros T l1.
NewInduction l1;Simpl;Auto.
Intro l2.
Rewrite (IHl1 l2).
Auto with zarith.
Qed.

Lemma Zlength_map : (T1,T2:Set) (l:(list T1)) (f:T1->T2) (Zlength T2 (map f l))=(Zlength T1 l).
Intros T1 T2 l.
NewInduction l;Simpl;Intros;Auto.
Rewrite (IHl f);Auto.
Qed.

Lemma Zlength_no_nil: (T:Set) (l:(list T)) (n:Z) `0<n` -> `(Zlength T l) = n` -> ~(l=(nil T)).
Intro T.
NewInduction l;Simpl;Intros.
Absurd  `0 = n`;Auto with zarith.
Discriminate.
Qed.

Lemma Zlength_le_0: (T:Set) (l:(list T)) `0<=(Zlength T l)`.
Intros.
NewInduction l;Simpl;Auto with zarith.
Qed.


(* ************************************************************************** *)
(* ****************************** Nth element ******************************* *)
(* ************************************************************************** *)
(** Nth element of a list *)

Inductive Znth_spec [T:Set] : (list T)->Z->T->Prop :=
  Znth_spec_O : (a:T)(l:(list T))(Znth_spec T (cons a l) `1` a)
| Znth_spec_S : (n:Z)(a,b:T)(l:(list T))
           (Znth_spec T l n a)->(Znth_spec T (cons b l) `n+1` a).

Hints Resolve Znth_spec_O Znth_spec_S. 

Definition Znth: (T:Set) (l:(list T)) (pos:Z) `1<=pos<=(Zlength T l)` -> T.
Intros T l.  
NewInduction l;Simpl;Intros.
Absurd `1 <= pos <= 0`;Auto.
Omega.
Case (Z_eq_dec pos `1`);Intros.
Assumption.
Assert `1 <= pos-1 <= (Zlength T l)`.
Omega.
Apply (IHl `pos-1` H0);Auto.
Defined.

Lemma Znth_exist : (T:Set) (l:(list T)) (pos:Z) `1<=pos<=(Zlength T l)` 
  -> (EX res:T | (Znth_spec T l pos res)).
Intros T l.  
NewInduction l;Simpl;Intros.
Absurd `1 <= pos <= 0`;Auto.
Omega.
Case (Z_eq_dec pos `1`);Intros.
Exists a.
Rewrite e.
Apply (Znth_spec_O T a l).
Assert `1 <= pos-1 <= (Zlength T l)`.
Omega.
Elim (IHl `pos-1` H0);Intros.
Exists x.
Generalize (Znth_spec_S T `pos-1` x a l H1).
Assert `pos-1+1`=`pos`.
Omega.
Rewrite H2;Auto.
Qed.



Axiom Znth_Znth_spec: (T:Set) (l:(list T)) (a:T) (n:Z) (Hyp_n:`1 <= n <= (Zlength T l)`) (a=(Znth T l n Hyp_n)) -> (Znth_spec T l n a).

Axiom Znth_ind : (T1,T2:Set) (l:(list T1)) (f:T1->T2) (n:Z) (H_n:`1<=n<=(Zlength T2 (map f l) )`)
                 (a:T1) (Znth_spec T1 l n a) -> ((Znth T2 (map f l) n H_n)=(f a)).


Lemma map_cons: (T1,T2:Set) (f:T1->T2) (a:T1) (l:(list T1)) (map f (cons a l))=(cons (f a) (map f l)).
Auto.
Qed.

Lemma Znth_lt_O : (A:Set) (l:(list A))(n:Z)(a:A)(Znth_spec A l n a)->`0<n`.
Intro.
Induction 1;Intros;Omega.
Qed. 


Lemma map_Znth:(T1,T2:Set) (l:(list T1)) (f:(T1->T2)) (pos:Z) (a:T1)
 (Znth_spec T1 l pos a) -> (Znth_spec T2 (map 1!T1 2!T2 f l) pos (f a)).
Intros T1 T2.
NewInduction l;Intros;Inversion H.
Rewrite (map_cons T1 T2 f a0 l).
Apply Znth_spec_O;Auto.
Rewrite (map_cons T1 T2 f a l).
Apply Znth_spec_S;Auto.
Qed.


Lemma map_Znth2:(T1,T2:Set) (l:(list T1)) (f:(T1->T2)) (a:T1) (b:T2) (b=(f a)) -> (pos:Z) 
 (Znth_spec T1 l pos a) -> (Znth_spec T2 (map 1!T1 2!T2 f l) pos b).
Intros T1 T2.
NewInduction l;Intros;Inversion H0.
Rewrite (map_cons T1 T2 f a0 l).
Rewrite H.
Apply Znth_spec_O;Auto.
Rewrite (map_cons T1 T2 f a l).
Apply Znth_spec_S;Auto.
Apply (IHl f a0 b H n H5).
Qed.

Lemma Znth_induc : (T:Set) (l:(list T)) (a,b:T) (n:Z) `0<n` -> (Znth_spec T (cons a l) `n+1` b) ->  (Znth_spec T l n b).
Intros.
Inversion H0.
Absurd `1 = n+1`;Auto with zarith.
Assert n=n0.
Auto with zarith.
Rewrite H6;Assumption.
Qed.

Lemma map_Znth3:(T1,T2:Set) (l:(list T1)) (f:(T1->T2)) (a:T1) (b:T2) (pos:Z) 
 (Znth_spec T1 l pos a) -> (Znth_spec T2 (map 1!T1 2!T2 f l) pos b) -> (b=(f a)).
Intros T1 T2.
Induction 1;Simpl;Intros.
Inversion H0;Auto.
Generalize (Znth_lt_O T2 (map f l0) n b H5);Intro.
Absurd `n+1 = 1`;Auto with zarith.
Generalize (Znth_lt_O T1 l0 n a0 H0);Intro.
Generalize (Znth_induc  T2 (map f l0) (f b0) b n  H3 H2);Auto.
Qed.

Lemma Znth_egal: (T:Set) (l:(list T)) (n:Z) (a,b:T) (Znth_spec T l n a) -> (Znth_spec T l n b) -> (a=b).
Induction 1;Simpl;Intros.
Inversion H0;Auto.
Generalize (Znth_lt_O T l0 n0 b H5);Intro.
Absurd `n0+1 = 1`;Auto with zarith.
Generalize (Znth_lt_O T l0 n0 a0 H0);Intro.
Generalize (Znth_induc  T l0 b0 b n0 H3 H2);Auto.
Qed.


Lemma if_Znth_then_In : (T:Set) (l:(list T)) (n:Z) `1 <= n <= (Zlength T l)` 
                         -> (a:T) (Znth_spec T l n a)
			 -> (In a l).
Intros T l n Hyp_n a Hyp_a.
NewInduction Hyp_a;Simpl;Intros;Auto.
Right.
Apply IHHyp_a.
Generalize (Znth_lt_O T l n a Hyp_a);Intro.
Generalize Hyp_n;Simpl;Intro.
Omega.
Qed.


Lemma Znth_app: (i:Z) (l:(list Z)) (Znth_spec Z l `i+1` i) -> (l':(list Z)) (Znth_spec Z (app l l') `i+1` i).
Intros i l H.
NewInduction H;Simpl;Intros.
Apply Znth_spec_O.
Generalize (IHZnth_spec l');Intro.
Apply Znth_spec_S;Auto.
Qed.

Lemma Znth_app2:  (l:(list Z)) (p:Z) ((Zlength Z l)=`p`) -> (a:Z) (Znth_spec Z (app l (cons a (nil Z))) `p+1` a).
Intros l.
NewInduction l;Simpl;Intros.
Rewrite (egal_reflexive ? ? ? H);Simpl.
Apply Znth_spec_O.

Assert `(Zlength Z l) = p-1`.
Omega.
Generalize (IHl `p-1` H0 a0).
Assert `p-1+1=p`.
Auto with zarith.
Rewrite H1.
Apply (Znth_spec_S Z `p` a0 a  (app l (cons a0 (nil Z))) ).
Qed.


(* ************************************************************************** *)
(* ****************************** From_to *********************************** *)
(* ************************************************************************** *)

Inductive from_to_spec  : (n:Z) (list Z) -> Prop :=
   from_to_0 :   (from_to_spec `0` (cons `0` (nil Z)))
 | from_to_rec : (p:Z) (l:(list Z)) (from_to_spec p l) -> (from_to_spec `p+1` (app l (cons `p+1` (nil Z)))).

Definition from_to_type := [n:Z] `0<=n` -> {res :(list Z) | (from_to_spec n res)}.

Definition from_to_F : (n:Z) ((y:Z)(Zwf `0` y n)->(from_to_type y))-> (from_to_type n).
Unfold 2 from_to_type.
Intros n fromto_recf H_n.
Refine ( if (Z_eq_dec `0` n) then [H1] (exist ? ? (cons `0` (nil Z)) ?)
           else [H2] (Cases (fromto_recf `n-1` ? ?) of
	   	   (exist l H) => (exist ? ? (app l (cons `n` (nil Z))) ?) end)).

Elim H1.
Apply from_to_0.
Unfold Zwf;Omega.
Auto with zarith.
Generalize (from_to_rec `n-1` l H).
Assert `n-1+1`=`n`.
Omega.
Rewrite H0;Auto.
Defined.

Definition from_to: (n:Z) `0<=n` -> {res :(list Z) | (from_to_spec n res)}
 := (well_founded_induction Z (Zwf `0`) (Zwf_well_founded `0`) (from_to_type ) (from_to_F )).


Lemma from_to_Zlength : (n:Z) (l:(list Z))  (h:(from_to_spec n l)) `(Zlength Z l) = n+1`.
Intros.
NewInduction h;Simpl;Auto.
Rewrite (Zlength_app Z l (cons (`p+1`) (nil Z)));Simpl;Auto with zarith.
Qed.

Lemma from_to_spec_not_inf : (p:Z) (l:(list Z)) (from_to_spec p l) -> `0<=p`.
Intros.
NewInduction H;Omega.
Qed.

Lemma from_to_Nth : (n:Z) `0<=n` -> (l:(list Z))  (h:(from_to_spec n l))
 (i:Z) `0<=i<=n` -> (Znth_spec Z l `i+1` i).
Intros n H_n l.
Induction 1.
Intros.
Simpl.
Assert i=`0`.
Auto with zarith.
Rewrite H0;Simpl.
Apply Znth_spec_O.
Intros.
Assert {`0<=i<= p`}+{`i=p+1`}.
Case (Z_ge_dec p i);Intros.
Left;Omega.
Right;Omega.
Case H2;Intros.
Generalize (H0 i a);Intro.
Apply Znth_app;Auto.
Assert `0 <= p <= p`.
Intuition.
Apply (from_to_spec_not_inf p l0 H).
Generalize (H0 p H3);Intro.
Rewrite e.
Assert `p+1+1=p+2`.
Auto with zarith.
Rewrite H5.
Generalize (from_to_Zlength p l0 H);Intro.
Generalize (Znth_app2 l0 `p+1` H6 `p+1`).
Assert `p+1+1=p+2`.
Auto with zarith.
Rewrite H7;Auto.
Qed.


Definition from_to2: (n:Z) (list Z).
Intro n.
Case (from_to (Zabs n) (Zabs_pos n));Auto.
Defined.

Lemma from_to2_nth: (i:Z) `0<=i` -> (pos:Z) `0 <= pos <= i` -> (Znth_spec Z (from_to2 i) `pos+1` pos).
Intros.
Unfold from_to2.
NewInduction (from_to `|i|` (Zabs_pos i)). 
Generalize (Zabs_eq i H);Intros.
Generalize p.
Rewrite H1;Intro.
Apply (from_to_Nth i H x p0 pos H0).
Qed.

Lemma from_to2_Zlength: (i:Z) `0<=i` -> (Zlength Z (from_to2 i))=`i+1`.
Intros.
Unfold from_to2.
NewInduction (from_to `|i|` (Zabs_pos i)).
Generalize (Zabs_eq i H);Intros.
Generalize p.
Rewrite H0;Intro.
Apply (from_to_Zlength i x p0 ).
Qed.


(* ************************************************************************** *)
(* ****************************** To_list  ********************************** *)
(* ************************************************************************** *)
(** ** To_list *)

Inductive to_list_spec [T1,T2:Set] [f:T1->(option T2)] : (list T1)->(list T2)->Prop := 
     to_list_nil  : (to_list_spec T1 T2 f (nil T1) (nil T2))
   | to_list_rec1 : (h:T1) (t:(list T1))  (res_rec:(list T2)) (to_list_spec T1 T2 f t res_rec)-> (x:T2) ((f h)=(Some T2 x)) -> (to_list_spec T1 T2 f (cons h t) (cons x res_rec))
   | to_list_rec2 : (h:T1) (t:(list T1))  (res_rec:(list T2)) (to_list_spec T1 T2 f t res_rec)-> ((f h)=(None T2)) -> (to_list_spec T1 T2 f (cons h t) res_rec).

Definition to_list : (T1,T2:Set)(l:(list T1))(f:T1->(option T2)) { res:(list T2) | (to_list_spec T1 T2 f l res)}.
Intros T1 T2 l.
NewInduction l;Intros.
Exists (nil T2).
Exact (to_list_nil T1 T2 f).
Elim (IHl f);Intros.
Case (option_dec T2 (f a));Intros.
Elim s;Intros.
Exists (cons x0 x).
Exact (to_list_rec1 T1 T2 f a l x p x0 p0).
Exists x.
Exact (to_list_rec2 T1 T2 f a l x p e).
Defined.


(* ************************************************************************** *)
(* ****************************** Filter_map ******************************** *)
(* ************************************************************************** *)

(** ** Filter_map *)
Inductive filtermap_spec [T1,T2:Set] [p:T1->bool] [f:T1->T2] : (list T1)->(list T2)->Prop := 
   filtermap_nil : (filtermap_spec T1 T2 p f (nil T1) (nil T2))
 | filtermap_rec1:  (h:T1) (t:(list T1)) (res_rec:(list T2)) (filtermap_spec T1 T2 p f t res_rec)-> ((p h)=true) -> (filtermap_spec T1 T2 p f (cons h t) (cons (f h) res_rec))
 | filtermap_rec2:  (h:T1) (t:(list T1)) (res_rec:(list T2)) (filtermap_spec T1 T2 p f t res_rec)-> ((p h)=false) -> (filtermap_spec T1 T2 p f (cons h t)  res_rec).

Definition filtermap : (T1,T2:Set)(l:(list T1))(p:T1->bool)(f:(T1->T2)) { res:(list T2) | (filtermap_spec T1 T2 p f l res)}.
Intros T1 T2 l.
NewInduction l;Intros.
Exists (nil T2).
Exact (filtermap_nil T1 T2 p f).
Elim (IHl p f);Intros.
Case (sumbool_of_bool (p a));Intros.
Exists (cons (f a) x).
Exact (filtermap_rec1 T1 T2 p f a l x p0 e).
Exists x.
Exact (filtermap_rec2 T1 T2 p f a l x p0 e).
Defined.
